@@ -2,14 +2,16 @@
 import json
 import os
 import re
+import tempfile
+import posixpath
 
 import logging
 from functools import wraps
 
-from flask import Response, request, send_file
+from flask import Response, request, current_app, send_file
 from google.protobuf import descriptor
 
-from mlflow.entities import Metric, Param, RunTag, ViewType, ExperimentTag
+from mlflow.entities import Metric, Param, RunTag, ViewType, ExperimentTag, FileInfo
 from mlflow.entities.model_registry import RegisteredModelTag, ModelVersionTag
 from mlflow.exceptions import MlflowException
 from mlflow.models import Model
@@ -63,6 +65,12 @@ from mlflow.protos.model_registry_pb2 import (
     SetModelVersionTag,
     DeleteModelVersionTag,
 )
+from mlflow.protos.mlflow_artifacts_pb2 import (
+    MlflowArtifactsService,
+    DownloadArtifact,
+    UploadArtifact,
+    ListArtifacts as ListArtifactsMlflowArtifacts,
+)
 from mlflow.protos.databricks_pb2 import RESOURCE_DOES_NOT_EXIST, INVALID_PARAMETER_VALUE
 from mlflow.store.artifact.artifact_repository_registry import get_artifact_repository
 from mlflow.store.db.db_types import DATABASE_ENGINES
@@ -76,6 +84,7 @@ from mlflow.tracking.registry import UnsupportedModelRegistryStoreURIException
 _logger = logging.getLogger(__name__)
 _tracking_store = None
 _model_registry_store = None
+_artifact_repo = None
 STATIC_PREFIX_ENV_VAR = "_MLFLOW_STATIC_PREFIX"
 
 
@@ -120,6 +129,21 @@ _tracking_store_registry = TrackingStoreRegistryWrapper()
 _model_registry_store_registry = ModelRegistryStoreRegistryWrapper()
 
 
+def _get_artifact_repo_mlflow_artifacts():
+    """
+    Get an artifact repository specified by `--artifacts-destination` option for `mlflow server`
+    command.
+    """
+    from mlflow.server import ARTIFACTS_DESTINATION_ENV_VAR
+
+    global _artifact_repo
+
+    if _artifact_repo is None:
+        _artifact_repo = get_artifact_repository(os.environ[ARTIFACTS_DESTINATION_ENV_VAR])
+
+    return _artifact_repo
+
+
 def _get_tracking_store(backend_store_uri=None, default_artifact_root=None):
     from mlflow.server import BACKEND_STORE_URI_ENV_VAR, ARTIFACT_ROOT_ENV_VAR
 
@@ -238,6 +262,44 @@ _TEXT_EXTENSIONS = [
 ]
 
 
+def _disable_unless_serve_artifacts(func):
+    @wraps(func)
+    def wrapper(*args, **kwargs):
+        from mlflow.server import SERVE_ARTIFACTS_ENV_VAR
+
+        if not os.environ.get(SERVE_ARTIFACTS_ENV_VAR):
+            return Response(
+                (
+                    f"Endpoint: {request.url_rule} disabled due to the mlflow server running "
+                    "without `--serve-artifacts`. To enable artifacts server functionality, "
+                    "run `mlflow server` with `--serve-artifacts`"
+                ),
+                503,
+            )
+        return func(*args, **kwargs)
+
+    return wrapper
+
+
+def _disable_if_artifacts_only(func):
+    @wraps(func)
+    def wrapper(*args, **kwargs):
+        from mlflow.server import ARTIFACTS_ONLY_ENV_VAR
+
+        if os.environ.get(ARTIFACTS_ONLY_ENV_VAR):
+            return Response(
+                (
+                    f"Endpoint: {request.url_rule} disabled due to the mlflow server running "
+                    "in `--artifacts-only` mode. To enable tracking server functionality, run "
+                    "`mlflow server` without `--artifacts-only`"
+                ),
+                503,
+            )
+        return func(*args, **kwargs)
+
+    return wrapper
+
+
 @catch_mlflow_exception
 def get_artifact_handler():
     from querystring_parser import parser
@@ -255,7 +317,11 @@ def _not_implemented():
     return response
 
 
+# Tracking Server APIs
+
+
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _create_experiment():
     request_message = _get_request_message(CreateExperiment())
     tags = [ExperimentTag(tag.key, tag.value) for tag in request_message.tags]
@@ -270,6 +336,7 @@ def _create_experiment():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _get_experiment():
     request_message = _get_request_message(GetExperiment())
     response_message = GetExperiment.Response()
@@ -281,6 +348,7 @@ def _get_experiment():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _get_experiment_by_name():
     request_message = _get_request_message(GetExperimentByName())
     response_message = GetExperimentByName.Response()
@@ -298,6 +366,7 @@ def _get_experiment_by_name():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _delete_experiment():
     request_message = _get_request_message(DeleteExperiment())
     _get_tracking_store().delete_experiment(request_message.experiment_id)
@@ -308,6 +377,7 @@ def _delete_experiment():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _restore_experiment():
     request_message = _get_request_message(RestoreExperiment())
     _get_tracking_store().restore_experiment(request_message.experiment_id)
@@ -318,6 +388,7 @@ def _restore_experiment():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _update_experiment():
     request_message = _get_request_message(UpdateExperiment())
     if request_message.new_name:
@@ -331,6 +402,7 @@ def _update_experiment():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _create_run():
     request_message = _get_request_message(CreateRun())
 
@@ -350,6 +422,7 @@ def _create_run():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _update_run():
     request_message = _get_request_message(UpdateRun())
     run_id = request_message.run_id or request_message.run_uuid
@@ -363,6 +436,7 @@ def _update_run():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _delete_run():
     request_message = _get_request_message(DeleteRun())
     _get_tracking_store().delete_run(request_message.run_id)
@@ -373,6 +447,7 @@ def _delete_run():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _restore_run():
     request_message = _get_request_message(RestoreRun())
     _get_tracking_store().restore_run(request_message.run_id)
@@ -383,6 +458,7 @@ def _restore_run():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _log_metric():
     request_message = _get_request_message(LogMetric())
     metric = Metric(
@@ -397,6 +473,7 @@ def _log_metric():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _log_param():
     request_message = _get_request_message(LogParam())
     param = Param(request_message.key, request_message.value)
@@ -409,6 +486,7 @@ def _log_param():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _set_experiment_tag():
     request_message = _get_request_message(SetExperimentTag())
     tag = ExperimentTag(request_message.key, request_message.value)
@@ -420,6 +498,7 @@ def _set_experiment_tag():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _set_tag():
     request_message = _get_request_message(SetTag())
     tag = RunTag(request_message.key, request_message.value)
@@ -432,6 +511,7 @@ def _set_tag():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _delete_tag():
     request_message = _get_request_message(DeleteTag())
     _get_tracking_store().delete_tag(request_message.run_id, request_message.key)
@@ -442,6 +522,7 @@ def _delete_tag():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _get_run():
     request_message = _get_request_message(GetRun())
     response_message = GetRun.Response()
@@ -453,6 +534,7 @@ def _get_run():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _search_runs():
     request_message = _get_request_message(SearchRuns())
     response_message = SearchRuns.Response()
@@ -476,6 +558,7 @@ def _search_runs():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _list_artifacts():
     request_message = _get_request_message(ListArtifacts())
     response_message = ListArtifacts.Response()
@@ -494,6 +577,7 @@ def _list_artifacts():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _get_metric_history():
     request_message = _get_request_message(GetMetricHistory())
     response_message = GetMetricHistory.Response()
@@ -506,6 +590,7 @@ def _get_metric_history():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _list_experiments():
     request_message = _get_request_message(ListExperiments())
     # `ListFields` returns a list of (FieldDescriptor, value) tuples for *present* fields:
@@ -523,11 +608,13 @@ def _list_experiments():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _get_artifact_repo(run):
     return get_artifact_repository(run.info.artifact_uri)
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _log_batch():
     _validate_batch_log_api_req(_get_request_json())
     request_message = _get_request_message(LogBatch())
@@ -544,11 +631,12 @@ def _log_batch():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _log_model():
     request_message = _get_request_message(LogModel())
     try:
         model = json.loads(request_message.model_json)
-    except:  # NB: can not be more specific here due to python2 compatibility
+    except Exception:
         raise MlflowException(
             "Malformed model info. \n {} \n is not a valid JSON.".format(
                 request_message.model_json
@@ -579,7 +667,11 @@ def _wrap_response(response_message):
     return response
 
 
+# Model Registry APIs
+
+
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _create_registered_model():
     request_message = _get_request_message(CreateRegisteredModel())
     registered_model = _get_model_registry_store().create_registered_model(
@@ -592,6 +684,7 @@ def _create_registered_model():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _get_registered_model():
     request_message = _get_request_message(GetRegisteredModel())
     registered_model = _get_model_registry_store().get_registered_model(name=request_message.name)
@@ -600,6 +693,7 @@ def _get_registered_model():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _update_registered_model():
     request_message = _get_request_message(UpdateRegisteredModel())
     name = request_message.name
@@ -612,6 +706,7 @@ def _update_registered_model():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _rename_registered_model():
     request_message = _get_request_message(RenameRegisteredModel())
     name = request_message.name
@@ -624,6 +719,7 @@ def _rename_registered_model():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _delete_registered_model():
     request_message = _get_request_message(DeleteRegisteredModel())
     _get_model_registry_store().delete_registered_model(name=request_message.name)
@@ -631,6 +727,7 @@ def _delete_registered_model():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _list_registered_models():
     request_message = _get_request_message(ListRegisteredModels())
     registered_models = _get_model_registry_store().list_registered_models(
@@ -644,6 +741,7 @@ def _list_registered_models():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _search_registered_models():
     request_message = _get_request_message(SearchRegisteredModels())
     store = _get_model_registry_store()
@@ -661,6 +759,7 @@ def _search_registered_models():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _get_latest_versions():
     request_message = _get_request_message(GetLatestVersions())
     latest_versions = _get_model_registry_store().get_latest_versions(
@@ -672,6 +771,7 @@ def _get_latest_versions():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _set_registered_model_tag():
     request_message = _get_request_message(SetRegisteredModelTag())
     tag = RegisteredModelTag(key=request_message.key, value=request_message.value)
@@ -680,6 +780,7 @@ def _set_registered_model_tag():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _delete_registered_model_tag():
     request_message = _get_request_message(DeleteRegisteredModelTag())
     _get_model_registry_store().delete_registered_model_tag(
@@ -689,6 +790,7 @@ def _delete_registered_model_tag():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _create_model_version():
     request_message = _get_request_message(CreateModelVersion())
     model_version = _get_model_registry_store().create_model_version(
@@ -704,6 +806,7 @@ def _create_model_version():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def get_model_version_artifact_handler():
     from querystring_parser import parser
 
@@ -716,6 +819,7 @@ def get_model_version_artifact_handler():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _get_model_version():
     request_message = _get_request_message(GetModelVersion())
     model_version = _get_model_registry_store().get_model_version(
@@ -727,6 +831,7 @@ def _get_model_version():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _update_model_version():
     request_message = _get_request_message(UpdateModelVersion())
     new_description = None
@@ -739,6 +844,7 @@ def _update_model_version():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _transition_stage():
     request_message = _get_request_message(TransitionModelVersionStage())
     model_version = _get_model_registry_store().transition_model_version_stage(
@@ -753,6 +859,7 @@ def _transition_stage():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _delete_model_version():
     request_message = _get_request_message(DeleteModelVersion())
     _get_model_registry_store().delete_model_version(
@@ -762,6 +869,7 @@ def _delete_model_version():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _get_model_version_download_uri():
     request_message = _get_request_message(GetModelVersionDownloadUri())
     download_uri = _get_model_registry_store().get_model_version_download_uri(
@@ -772,6 +880,7 @@ def _get_model_version_download_uri():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _search_model_versions():
     request_message = _get_request_message(SearchModelVersions())
     model_versions = _get_model_registry_store().search_model_versions(request_message.filter)
@@ -781,6 +890,7 @@ def _search_model_versions():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _set_model_version_tag():
     request_message = _get_request_message(SetModelVersionTag())
     tag = ModelVersionTag(key=request_message.key, value=request_message.value)
@@ -791,6 +901,7 @@ def _set_model_version_tag():
 
 
 @catch_mlflow_exception
+@_disable_if_artifacts_only
 def _delete_model_version_tag():
     request_message = _get_request_message(DeleteModelVersionTag())
     _get_model_registry_store().delete_model_version_tag(
@@ -799,6 +910,82 @@ def _delete_model_version_tag():
     return _wrap_response(DeleteModelVersionTag.Response())
 
 
+# MLflow Artifacts APIs
+
+
+@catch_mlflow_exception
+@_disable_unless_serve_artifacts
+def _download_artifact(artifact_path):
+    """
+    A request handler for `GET /mlflow-artifacts/artifacts/<artifact_path>` to download an artifact
+    from `artifact_path` (a relative path from the root artifact directory).
+    """
+    basename = posixpath.basename(artifact_path)
+    tmp_dir = tempfile.TemporaryDirectory()
+    tmp_path = os.path.join(tmp_dir.name, basename)
+    artifact_repo = _get_artifact_repo_mlflow_artifacts()
+    artifact_repo._download_file(artifact_path, tmp_path)
+
+    # Ref: https://stackoverflow.com/a/24613980/6943581
+    file_handle = open(tmp_path, "rb")
+
+    def stream_and_remove_file():
+        yield from file_handle
+        file_handle.close()
+        tmp_dir.cleanup()
+
+    return current_app.response_class(
+        stream_and_remove_file(),
+        headers={"Content-Disposition": "attachment", "filename": basename},
+    )
+
+
+@catch_mlflow_exception
+@_disable_unless_serve_artifacts
+def _upload_artifact(artifact_path):
+    """
+    A request handler for `PUT /mlflow-artifacts/artifacts/<artifact_path>` to upload an artifact
+    to `artifact_path` (a relative path from the root artifact directory).
+    """
+    head, tail = posixpath.split(artifact_path)
+    with tempfile.TemporaryDirectory() as tmp_dir:
+        tmp_path = os.path.join(tmp_dir, tail)
+        with open(tmp_path, "wb") as f:
+            chunk_size = 1024 * 1024  # 1 MB
+            while True:
+                chunk = request.stream.read(chunk_size)
+                if len(chunk) == 0:
+                    break
+                f.write(chunk)
+
+        artifact_repo = _get_artifact_repo_mlflow_artifacts()
+        artifact_repo.log_artifact(tmp_path, artifact_path=head or None)
+
+    return _wrap_response(UploadArtifact.Response())
+
+
+@catch_mlflow_exception
+@_disable_unless_serve_artifacts
+def _list_artifacts_mlflow_artifacts():
+    """
+    A request handler for `GET /mlflow-artifacts/artifacts?path=<value>` to list artifacts in `path`
+    (a relative path from the root artifact directory).
+    """
+    request_message = _get_request_message(ListArtifactsMlflowArtifacts())
+    path = request_message.path if request_message.HasField("path") else None
+    artifact_repo = _get_artifact_repo_mlflow_artifacts()
+    files = []
+    for file_info in artifact_repo.list_artifacts(path):
+        basename = posixpath.basename(file_info.path)
+        new_file_info = FileInfo(basename, file_info.is_dir, file_info.file_size)
+        files.append(new_file_info.to_proto())
+    response_message = ListArtifacts.Response()
+    response_message.files.extend(files)
+    response = Response(mimetype="application/json")
+    response.set_data(message_to_json(response_message))
+    return response
+
+
 def _add_static_prefix(route):
     prefix = os.environ.get(STATIC_PREFIX_ENV_VAR)
     if prefix:
@@ -838,7 +1025,11 @@ def get_endpoints():
                     ret.append((http_path, handler, [endpoint.method]))
         return ret
 
-    return get_service_endpoints(MlflowService) + get_service_endpoints(ModelRegistryService)
+    return (
+        get_service_endpoints(MlflowService)
+        + get_service_endpoints(ModelRegistryService)
+        + get_service_endpoints(MlflowArtifactsService)
+    )
 
 
 HANDLERS = {
@@ -885,4 +1076,8 @@ HANDLERS = {
     DeleteRegisteredModelTag: _delete_registered_model_tag,
     SetModelVersionTag: _set_model_version_tag,
     DeleteModelVersionTag: _delete_model_version_tag,
+    # MLflow Artifacts APIs
+    DownloadArtifact: _download_artifact,
+    UploadArtifact: _upload_artifact,
+    ListArtifactsMlflowArtifacts: _list_artifacts_mlflow_artifacts,
 }
